import unittest

from mittmcts import MCTS

from test.connect4 import (
    get_bitboards, find_row_for_column, empty_board, ConnectFourGame
)


class TestConnectFour(unittest.TestCase):
    def test_empty_board(self):
        _ = None
        self.assertEqual(empty_board(),
                         [[_, _, _, _, _, _, _],
                          [_, _, _, _, _, _, _],
                          [_, _, _, _, _, _, _],
                          [_, _, _, _, _, _, _],
                          [_, _, _, _, _, _, _],
                          [_, _, _, _, _, _, _]])

    def test_find_row_for_column(self):
        _ = None
        board = [[0, 0, _, _, _, 1, 1],
                 [1, 0, _, _, _, 0, 0],
                 [0, 1, 0, _, _, 1, 1],
                 [1, 0, 1, _, _, 0, 0],
                 [0, 1, 0, _, _, 1, 1],
                 [1, 0, 1, _, 1, 0, 0]]
        self.assertEqual(find_row_for_column(board, 2), 1)
        with self.assertRaises(Exception):
            find_row_for_column(board, 0)
        self.assertEqual(find_row_for_column(board, 3), 5)
        self.assertEqual(find_row_for_column(board, 4), 4)
        with self.assertRaises(Exception):
            find_row_for_column(board, 5)

    def test_apply_move(self):
        _ = None
        board = [[_, _, _, _, _, _, _],
                 [_, _, _, _, _, _, _],
                 [_, _, _, _, _, _, _],
                 [_, _, _, _, _, _, _],
                 [_, _, _, _, _, _, _],
                 [_, _, _, _, 0, 0, 0]]
        initial_state = ConnectFourGame.initial_state()
        initial_state = initial_state._replace(board=board,
                                               bitboards=get_bitboards(board))
        state = ConnectFourGame.apply_move(initial_state, 6)
        self.assertIsNone(state.winner)
        self.assertEqual(state.board,
                         [[_, _, _, _, _, _, _],
                          [_, _, _, _, _, _, _],
                          [_, _, _, _, _, _, _],
                          [_, _, _, _, _, _, _],
                          [_, _, _, _, _, _, 0],
                          [_, _, _, _, 0, 0, 0]])
        state = ConnectFourGame.apply_move(initial_state, 3)
        self.assertEqual(state.board,
                         [[_, _, _, _, _, _, _],
                          [_, _, _, _, _, _, _],
                          [_, _, _, _, _, _, _],
                          [_, _, _, _, _, _, _],
                          [_, _, _, _, _, _, _],
                          [_, _, _, 0, 0, 0, 0]])
        self.assertEqual(state.winner, 0)

        _ = None
        board = [[_, _, _, _, _, _, _],
                 [_, _, _, _, _, _, _],
                 [_, _, _, _, _, _, _],
                 [_, _, _, _, _, 1, 0],
                 [_, _, _, _, 1, 0, 1],
                 [_, _, _, 1, 0, 0, 0]]
        state = ConnectFourGame.initial_state()
        state = state._replace(board=board,
                               bitboards=get_bitboards(board),
                               current_player=1)
        state = ConnectFourGame.apply_move(state, 6)
        self.assertEqual(state.winner, 1)

    def test_with_mcts(self):
        _ = None
        board = [[_, _, _, _, _, _, _],
                 [_, _, _, _, _, _, _],
                 [_, _, _, _, _, _, _],
                 [_, _, _, _, _, _, _],
                 [_, _, 1, 1, 1, _, _],
                 [_, _, 0, 0, 0, _, _]]
        state = ConnectFourGame.State(
            board=board,
            current_player=0,
            bitboards=get_bitboards(board),
            winner=None)
        result = (MCTS(ConnectFourGame, state)
                  .get_simulation_result(100))
        self.assertIn(result.move, [1, 5])



from collections import defaultdict, namedtuple, Counter
from copy import deepcopy

from math import sqrt, log
from random import choice, random

from six import iteritems


class Draw(object):
    pass
Draw = Draw()


class Node(object):
    def __init__(self, game, state, parent, move, c, depth=0):
        self.parent = parent
        self.__state = state
        if parent is None:
            self.__initial_state = deepcopy(state)
        self.__children = {}
        self.game = game
        self.move = move
        self.visits = 0
        self.draws = 0
        self.wins_by_player = defaultdict(lambda: 0)
        self.misc_by_player = defaultdict(dict)
        self.impossible_state = False
        self.c = c
        self.depth = depth

    def ucb1(self, player):
        if not self.parent:
            return 0
        wins_by_player = self.wins_by_player.get(player, 0)
        try:
            ucb = (
                (float(wins_by_player + (self.draws * 0.5)) / self.visits) +
                (self.c * sqrt(log(self.parent.visits) / self.visits)))
        except ZeroDivisionError:
            ucb = 0
        return ucb

    @property
    def winner(self):
        return self.game.get_winner(self.state)

    @property
    def state(self):
        if not self.__state:
            self.__state = self.game.apply_move(self.parent.state, self.move)
        return self.__state

    def determine(self):
        # if games implement a determine classmethod then we are
        # doing ISMCTS so we randomly pick the hidden state every time
        # we play out
        if hasattr(self.game, 'determine'):
            self.__state = self.game.determine(self.__initial_state)

    def add_new_children_for_determination(self, moves):
        self.__children.update({move: Node(game=self.game,
                                           state=None,
                                           move=move,
                                           parent=self,
                                           c=self.c,
                                           depth=self.depth + 1)
                                for move in moves
                                if move not in self.__children})

    @property
    def children(self):
        is_random, moves = self.game.get_moves(self.state)
        self.is_random = is_random
        self.add_new_children_for_determination(moves)
        return [child for move, child in iteritems(self.__children)
                if move in moves]

    def get_best_child(self):
        # force instantiation of child nodes and get self.is_random set
        children = self.children

        if not children:
            raise ValueError('Need to have children to find the '
                             'best child')

        if self.is_random:
            return choice(children)

        # visit unplayed moves first
        # if all moves have been visited then visit the move with the highest
        # ucb1 payout
        children = sorted(children,
                          key=lambda c: (c.visits == 0 and random() or -1,
                                         c.ucb1(self.current_player)))
        return children[-1]

    @property
    def current_player(self):
        return self.game.current_player(self.state)

    def dump_tree(self):
        print(repr(self))
        children = self.children
        while children:
            next_children = []
            for child in children:
                print(repr(child))
                next_children.extend(child.children)
            children = next_children

    def most_visited_child(self, actual_options=None):
        children = self.children
        if actual_options:
            self.add_new_children_for_determination(actual_options)
            children = [child for move, child in iteritems(self.__children)
                        if move in actual_options]
        if not children:
            raise Exception('No children when trying to find most visited '
                            'move\n actual_options=%r children=%r' %
                            (actual_options, children))

        return sorted(children, key=lambda c: c.visits)[-1]

    def backprop(self):
        winner = self.winner
        current_node = self
        update_misc = None
        if hasattr(self.game, 'update_misc'):
            update_misc = self.game.update_misc
        while current_node:
            current_node.visits += 1
            if winner is Draw:
                current_node.draws += 1
            else:
                current_node.wins_by_player[winner] += 1
            if update_misc:
                update_misc(self.state, current_node.misc_by_player)
            current_node = current_node.parent

    def reset_state(self):
        self.__state = None

    def __repr__(self):
        return 'state=%r move=%r visits=%r wins=%r ucb=%r' % (
            self.state,
            self.move,
            self.visits,
            self.wins_by_player[self.parent.current_player],
            self.ucb1(self.parent.current_player))


class MCTS(object):
    def __init__(self, game, initial_state=None, c=sqrt(2)):
        self.game = game
        self.c = c
        if initial_state:
            self.__initial_state = initial_state
        else:
            self.__initial_state = game.initial_state()

        self.__initial_state = deepcopy(self.__initial_state)

    def get_simulation_result(self,
                              iterations=1,
                              actual_options=None,
                              get_leaf_nodes=False):
        root_node = Node(game=self.game,
                         parent=None,
                         state=self.__initial_state,
                         move=None,
                         c=self.c)
        MCTSResult = namedtuple('MCTSResult', 'root, move, leaf_nodes,'
                                              'max_depth, avg_depth')
        plays = 0
        max_depth = 0
        total_depth = 0
        leaf_nodes = []
        determined = hasattr(self.game, 'determine')
        while plays < iterations:
            root_node.determine()
            current_node = root_node
            while current_node.winner is None and current_node.children:
                current_node = current_node.get_best_child()
                if determined:
                    current_node.reset_state()
            if current_node.winner is None:
                raise ValueError('A game cannot have a terminal node that '
                                 'has no winner. If the game was a draw '
                                 'return Draw')
            current_node.backprop()
            max_depth = max(max_depth, current_node.depth)
            total_depth += current_node.depth
            plays += 1
            if get_leaf_nodes:
                    leaf_nodes.append(current_node)

        move = root_node.most_visited_child(actual_options).move
        return MCTSResult(root=root_node,
                          move=move,
                          leaf_nodes=leaf_nodes,
                          avg_depth=float(total_depth) / plays,
                          max_depth=max_depth)


def flamegraph(mcts_result, depth=None):
    root_node = mcts_result.root
    leaf_nodes = mcts_result.leaf_nodes
    walks = Counter()
    current_player = root_node.current_player
    for node in leaf_nodes:
        moves = []
        while node:
            extra = ''
            if node == root_node:
                break
            else:
                move = node.move
            if node.parent.current_player != current_player:
                extra = '-opp'
            if node.winner == Draw:
                moves.insert(0, '{}-draw'.format(current_player))
            elif node.winner == current_player:
                moves.insert(0, '{}-win'.format(current_player))
            elif node.winner is not None:
                moves.insert(0, '{}-lose'.format(current_player))
            moves.insert(0, str(move) + extra)
            node = node.parent
        if depth is not None:
            moves = moves[:depth]
        walks[';'.join(moves)] += 1
    for path, count in walks.iteritems():
        print('{} {}'.format(path, count))



import unittest

from mock import patch

from test.games import (
    GameWithOneMove, GameWithTwoMoves, SimpleDiceRollingGame, TicTacToeGame,
    GameWithManyMovesOnlyOneDetermined
)
from mittmcts import MCTS, Draw


class TestMCTS(unittest.TestCase):
    def test_game_with_one_move(self):
        result = MCTS(GameWithOneMove).get_simulation_result(100)
        self.assertEqual(result.move, 'win')
        self.assertEqual(result.root.children[0].wins_by_player[1], 100)
        self.assertEqual(result.max_depth, 1)

    def test_game_with_two_possible_moves(self):
        result = MCTS(GameWithTwoMoves).get_simulation_result(100)
        self.assertEqual(result.root.children[0].move, 0)
        self.assertEqual(result.root.children[0].wins_by_player[1], 0)
        self.assertIsNone(result.root.children[0].winner)
        self.assertEqual(result.root.children[0].children[0].winner, 2)
        self.assertEqual(result.root.children[0].children[0].wins_by_player,
                         {2: result.root.children[0].children[0].visits})
        self.assertEqual(result.root.children[1].move, 1)
        self.assertEqual(result.root.children[1].winner, 1)
        self.assertEqual(result.root.children[1].wins_by_player,
                         {1: result.root.children[1].visits})
        self.assertEqual(result.move, 1)
        self.assertEqual(result.max_depth, 2)
        self.assertEqual(result.avg_depth, 1.06)

    def test_random_moves_selected_randomly(self):
        with patch('mittmcts.choice') as mock_choice:
            # always choose the first item in random choices
            # (lowest die rolls in our silly game)
            mock_choice.side_effect = lambda items: items[0]
            result = MCTS(SimpleDiceRollingGame).get_simulation_result(100)
            # because the player chooes to roll no dice it always loses
            self.assertEqual(result.root.children[0].wins_by_player[1], 0)
            self.assertEqual(result.root.children[1].wins_by_player[1], 0)
            self.assertEqual(result.root.children[2].wins_by_player[1], 0)
            self.assertEqual(result.root.misc_by_player[1]['min_score'], 0)
            self.assertEqual(result.root.misc_by_player[1]['max_score'], 2)
            self.assertEqual(result.root.children[1]
                             .misc_by_player[1]['min_score'],
                             1)
            self.assertEqual(result.root.children[1]
                             .misc_by_player[1]['max_score'],
                             1)
            self.assertEqual(result.root.children[2]
                             .misc_by_player[1]['min_score'],
                             2)
            self.assertEqual(result.root.children[2]
                             .misc_by_player[1]['max_score'],
                             2)
            self.assertEqual(result.root.wins_by_player[1], 0)

        with patch('mittmcts.choice') as mock_choice:
            mock_choice.side_effect = lambda items: items[-1]
            result = MCTS(SimpleDiceRollingGame).get_simulation_result(100)
            # 100 simulations should be enough time for UCB1 to converge on
            # the always winning choice given our loaded dice
            self.assertEqual(result.move, 2)
            # because the player chooses the highest die roll the player wins
            # every time it rolls two dice
            self.assertEqual(result.root.children[2].wins_by_player[1],
                             result.root.children[2].visits)
            self.assertEqual(result.root.children[2]
                             .misc_by_player[1]['min_score'],
                             12)
            self.assertEqual(result.root.children[2]
                             .misc_by_player[1]['max_score'],
                             12)
            self.assertEqual(result.root.children[2]
                             .misc_by_player[1]['avg_score'],
                             12)

    def test_selects_winning_tictactoe_move(self):
        ___ = None
        one_move_from_winning = TicTacToeGame.State(board=['O', 'O', ___,
                                                           'X', ___, 'X',
                                                           ___, 'X', ___],
                                                    current_player='O',
                                                    winner=None)
        result = (MCTS(TicTacToeGame, one_move_from_winning)
                  .get_simulation_result(100))
        self.assertEqual(result.move, 2)
        one_move_from_winning = TicTacToeGame.State(board=['O', ___, ___,
                                                           'O', 'X', 'X',
                                                           ___, 'X', ___],
                                                    current_player='O',
                                                    winner=None)
        result = (MCTS(TicTacToeGame, one_move_from_winning)
                  .get_simulation_result(100))
        self.assertEqual(result.move, 6)

    def test_tictactoe_tie(self):
        ___ = None
        one_move_from_winning = TicTacToeGame.State(board=['O', 'X', ___,
                                                           'X', 'O', 'O',
                                                           'O', 'X', 'X'],
                                                    current_player='X',
                                                    winner=None)
        result = (MCTS(TicTacToeGame, one_move_from_winning)
                  .get_simulation_result(100))
        self.assertEqual(result.move, 2)
        self.assertEqual(result.root.children[0].winner, Draw)

    def test_only_determined_moves_are_followed(self):
        result = (MCTS(GameWithManyMovesOnlyOneDetermined)
                  .get_simulation_result(100))
        self.assertEqual(result.root.children[0].move, 1)
        self.assertEqual(result.root.children[0].visits, 100)



#!/usr/bin/env python

from distutils.core import setup

setup(
    name='mittmcts',
    version='0.2',
    description='Man in the Table Information Set Monte Carlo Tree Search '
                'library',
    author='Dan Bravender',
    author_email='dan@bravender.net',
    url='https://github.com/dbravender/mittmcts',
    packages=['mittmcts'],
)



from itertools import chain
import unittest

from mittmcts import MCTS

from test.euchre import (
    second_highest_jack, winning_card, deal, sort_by_trump_and_lead,
    playable_cards, suit, potential_cards_given_voids,
    EuchreGame
)


class TestEuchre(unittest.TestCase):
    def test_deal(self):
        self.assertEqual(deal(),
                         ['ad', 'kd', 'qd', 'jd', '0d', '9d',
                          'ah', 'kh', 'qh', 'jh', '0h', '9h',
                          'as', 'ks', 'qs', 'js', '0s', '9s',
                          'ac', 'kc', 'qc', 'jc', '0c', '9c'])

    def test_second_highest_jack(self):
        self.assertEqual(second_highest_jack('c'), 'js')
        self.assertEqual(second_highest_jack('s'), 'jc')
        self.assertEqual(second_highest_jack('d'), 'jh')
        self.assertEqual(second_highest_jack('h'), 'jd')

    def test_suit(self):
        self.assertEqual(suit('s', 'jc'), 's')
        self.assertEqual(suit('c', 'js'), 'c')
        self.assertEqual(suit('d', 'jh'), 'd')
        self.assertEqual(suit('h', 'jd'), 'h')

    def test_sort_by_trump_and_lead(self):
        self.assertEqual(sort_by_trump_and_lead('d', 'h', deal())[:12],
                         ['jd', 'jh', 'ad', 'kd', 'qd', '0d', '9d',
                          'ah', 'kh', 'qh', '0h', '9h'])
        self.assertEqual(sort_by_trump_and_lead('c', 'd', deal())[:13],
                         ['jc', 'js', 'ac', 'kc', 'qc', '0c', '9c',
                          'ad', 'kd', 'qd', 'jd', '0d', '9d'])

    def test_winning_card(self):
        self.assertEqual(winning_card('c', 'c', ['jc', 'js', 'as', '0s']),
                         'jc')
        self.assertEqual(winning_card('c', 'c', ['js', 'ac', 'jd', '0s']),
                         'js')
        self.assertEqual(winning_card('d', 'c', ['jc', 'ac', '0s', '9s']),
                         'ac')
        self.assertEqual(winning_card('d', 'c', ['9d', 'ac', '0s', '9s']),
                         '9d')

    def test_playable_cards(self):
        self.assertEqual(playable_cards('c', 'd', deal()),
                         ['ad', 'kd', 'qd', 'jd', '0d', '9d'])
        self.assertEqual(playable_cards('c', 'c', deal()),
                         ['js', 'ac', 'kc', 'qc', 'jc', '0c', '9c'])
        self.assertEqual(playable_cards('c', 'c', deal()),
                         ['js', 'ac', 'kc', 'qc', 'jc', '0c', '9c'])
        self.assertEqual(playable_cards('c', 'c', ['ad']),
                         ['ad'])

    def test_potential_cards_given_voids(self):
        self.assertEqual(
            potential_cards_given_voids('c', ['d'], ['jd', 'jc']), ['jc'])
        self.assertEqual(
            potential_cards_given_voids('d', ['d'], ['jd', 'jh']), [])

    def test_initial_state_bad_cards(self):
        try:
            EuchreGame.initial_state(
                ['notacard', 'notacard2', 'notacard4', 'notacard5'])
            self.fail('Should not allow invalid starting hands')
        except ValueError:
            pass

    def test_initial_state_set_trump_and_determine_deals(self):
        state = EuchreGame.initial_state(['ad', 'kd', 'qd', 'jd', '0d'])
        state = EuchreGame.determine(state)
        self.assertEqual(len(list(chain(*state.hands))), 20)
        state = EuchreGame.initial_state(trump_card='jc')
        self.assertEqual(state.trump, 'c')

    def test_initial_state_invalid_trump(self):
        try:
            EuchreGame.initial_state(trump_card='xp')
            self.fail('initial_state should require valid trump')
        except ValueError:
            pass

    def test_initial_state_trump_determined_by_kitty(self):
        state = EuchreGame.initial_state()
        state = EuchreGame.determine(state)
        self.assertEqual(len(state.cards_played), 1)
        self.assertEqual(suit(state.trump, state.cards_played[0]), state.trump)
        self.assertIn(state.trump_card, state.cards_played)

    def test_apply_move_sets_a_winner_after_4_plays(self):
        state = EuchreGame.initial_state()
        state = EuchreGame.determine(state)
        for x in range(4):
            lead_suit = suit(state.trump, state.lead_card)
            move = playable_cards(state.trump,
                                  lead_suit,
                                  state.hands[state.current_player], )[0]
            state = EuchreGame.apply_move(state, move)
        self.assertEqual(sum(state.tricks_won_by_team), 1)

    def test_determine(self):
        state = EuchreGame.initial_state()
        state = state._replace(trump_card='jd',
                               trump='d',
                               cards_played=['jd', 'ad', 'kd', 'qd', 'qc'],
                               tricks_won_by_team=[1, 0],
                               hands=[['jc', 'kc', 'ah', 'js'],
                                      [],
                                      [],
                                      []],
                               voids_by_player=[set(),
                                                set(['d', 'h', 'c']),
                                                set(),
                                                set(['s', 'c', 'd'])])
        state = EuchreGame.determine(state)
        self.assertTrue(all([suit('d', card) == 's'
                             for card in state.hands[1]]))
        self.assertTrue(all([suit('d', card) == 'h'
                             for card in state.hands[3]]))

    def test_determine_in_the_middle_of_a_trick(self):
        state = EuchreGame.initial_state()
        state = state._replace(trump_card='jd',
                               trump='d',
                               tricks_won_by_team=[1, 0],
                               cards_played_by_player=['jh', None, None, None],
                               hands=[[],
                                      [],
                                      [],
                                      []],
                               voids_by_player=[set(),
                                                set(['d', 'h', 'c']),
                                                set(),
                                                set(['s', 'c', 'd'])])
        state = EuchreGame.determine(state)
        self.assertEqual(len(state.hands[0]), 3)
        self.assertEqual(len(state.hands[1]), 4)
        self.assertEqual(len(state.hands[2]), 4)
        self.assertEqual(len(state.hands[3]), 4)

    def test_a_whole_hand(self):
        state = EuchreGame.initial_state(['jd', 'jh', '9c', '9h', 'as'], 'kd')
        state = state._replace(hands=[['jd', 'jh', '9c', '9h', 'as'],
                                      ['qd', 'ah', '0h', 'kc', 'js'],
                                      ['9d', 'jc', 'kh', 'qc', '9s'],
                                      ['0d', '0c', 'qh', 'ac', '0s']])
        state = EuchreGame.apply_move(state, 'jd')
        state = EuchreGame.apply_move(state, 'qd')
        state = EuchreGame.apply_move(state, '9d')
        state = EuchreGame.apply_move(state, '0d')
        self.assertEqual(state.tricks_won_by_team, [1, 0])
        self.assertEqual(state.current_player, 0)
        state = EuchreGame.apply_move(state, 'jh')
        state = EuchreGame.apply_move(state, 'ah')
        state = EuchreGame.apply_move(state, 'jc')
        state = EuchreGame.apply_move(state, '0c')
        self.assertEqual(state.tricks_won_by_team, [2, 0])
        self.assertEqual(state.current_player, 0)
        state = EuchreGame.apply_move(state, '9h')
        state = EuchreGame.apply_move(state, '0h')
        state = EuchreGame.apply_move(state, 'kh')
        state = EuchreGame.apply_move(state, 'qh')
        self.assertEqual(state.tricks_won_by_team, [3, 0])
        self.assertEqual(state.current_player, 2)
        state = EuchreGame.apply_move(state, 'qc')
        state = EuchreGame.apply_move(state, 'ac')
        state = EuchreGame.apply_move(state, '9c')
        state = EuchreGame.apply_move(state, 'kc')
        self.assertEqual(state.tricks_won_by_team, [3, 1])
        self.assertEqual(state.current_player, 3)
        self.assertIsNone(state.winning_team)
        state = EuchreGame.apply_move(state, '0s')
        try:
            EuchreGame.apply_move(state, 'ad')
            self.fail('Did not follow suit earlier')
        except ValueError:
            pass
        state = EuchreGame.apply_move(state, 'as')
        state = EuchreGame.apply_move(state, 'js')
        state = EuchreGame.apply_move(state, '9s')
        self.assertEqual(state.tricks_won_by_team, [4, 1])
        self.assertEqual(state.winning_team, 0)
        self.assertEqual(EuchreGame.get_winner(state), 0)
        self.assertEqual(EuchreGame.current_player(state), 0)

    def test_with_mcts(self):
        state = EuchreGame.initial_state(['0d', '0h', 'as', 'ac', 'ah'], 'jd')
        result = MCTS(EuchreGame, state).get_simulation_result(100)
        self.assertEqual(result.max_depth, 20)
        # there is no longer option now so all players playing will always
        # result in 20 cards being played
        self.assertEqual(result.avg_depth, 20)

    def test_this_hand_should_win_every_time(self):
        state = EuchreGame.State(
            cards_played_by_player=[None, None, None, None],
            current_player=0,
            lead_card=None,
            trump_card='0d',
            trump='d',
            winning_team=None,
            hands=[['jd', 'jh', 'ad', 'kd', 'qd'], [], [], []],
            tricks_won_by_team=[0, 0],
            cards_played=['0d'],
            voids_by_player=[set([]), set([]), set([]), set([])])
        result = (MCTS(EuchreGame, state)
                  .get_simulation_result(100, get_leaf_nodes=True))
        for node in result.leaf_nodes:
            self.assertGreaterEqual(node.state.tricks_won_by_team[0], 5)

    def test_determine_deals_5_cards_to_each_player(self):
        state = EuchreGame.initial_state()
        state = EuchreGame.determine(state)
        self.assertTrue(all(len(hand) == 5 for hand in state.hands))
